%% Contraction of finite PEPS
% Author: <https://www.theorie.physik.uni-muenchen.de/lsvondelft/members/sci_mem/seung_sup-lee/index.html 
% Seung-Sup Lee>
%% 
% In this tutorial, we will compute the correlation functions of projected entangled-pair 
% states (PEPSs) on a finite-sized square lattice with open boundary condition. 
% For example, a ket PEPS on a $3 \times 4$ square lattice is represented by:
% 
% 
% 
% Here rank-5 tensors (blue balls) are arranged on a square grid of three rows 
% and four columns. We index the tensors by its row and column indices: the upper-left 
% corner is $(1,1)$ and the lower-right corner is $(3,4)$. We use the leg order 
% convention of left-up-physical-down-right. The right and down legs are outward, 
% and the rest are inward. The open legs at the boundary, which do not connect 
% tensors, are dummy legs.
% 
% The squared norm of the PEPS is obtained by the contraction of tensor network 
% that consists of the ket (upper layer) and bra (lower layer) PEPSs:
% 
% 
% 
% The spin-spin correlation function $\langle \hat{S}_{(i,j)}^z \hat{S}_{(i',j')}^z 
% \rangle$ is obtained by the contraction of the ket and bra layers of PEPS with 
% spin-$z$ operators acting on sites $(i,j)$ and $(i',j')$:
% 
% 
% 
% Here the green balls indicate the rank-2 form of spin-$z$ operators.
% 
% To contract such tensor networks, we use the following strategy. First, we 
% contract all the tensors associated with the same lattice site, and fuse the 
% in-plane legs by using isometries (yellow triangles).
% 
% 
% 
% The result is called double tensor or transfer operator. Similarly, for the 
% sites on which local spin operators act, we have
% 
% 
% 
% Then the tensor networks become purely two-dimensional. For example, the tensor 
% network with spin-$z$ operators, which is shown above, reduces to:
% 
% 
% 
% We contract this network column by column. The first (left-most) column is 
% equivalent to a matrix product state (MPS) since the left open legs of the column 
% are dummy legs. And the second and third columns are equivalent to matrix product 
% operators (MPOs). We first contract the first and second columns, then contract 
% the result with the third column, and so on.
% 
% 
% 
% After contracting with another column, we apply the downward and upward sweeps 
% of truncations: the bonds lying along column directions are truncated via the 
% singular value decomposition on the rank-4 tensor which is the contraction of 
% two nearest-neighbor tensors.
%% Exercise (a): Complete |normFinPEPS_Ex.m|
% There is a function |normFinPEPS_Ex.m| which is zipped together with this 
% tutorial material. This function is designed to perform the contraction of the 
% tensor networks made of the bra and ket PEPSs and, if given, local operators, 
% following the strategy described above.
% 
% In this function, we consider the ket PEPS is uniform in the bulk, i.e., is 
% maded of the same local tensor |A| as input. The tensors at the boundary are 
% obtained by projecting the space of the open legs onto their first bases. That 
% is, the bond space of local tensor |A| should be defined in a way that the first 
% basis corresponds to the vacuum space.
% 
% *Complete the parts which are enclosed by the comments |TODO - Exercise (a)|.* 
% Once you complete the function, you can follow the demonstration below.
%% Resonating valence bond (RVB) state
% As a paradigmatic PEPS, we consider the RVB state on a square lattice. In 
% this tutorial, we use the PEPS representation of the RVB state introduced in 
% <https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.96.220601 F. Verstraete, 
% M. M. Wolf, D. Perez-Garcia, and J. I. Cirac, Phys. Rev. Lett. *96*, 220601 
% (2006)> (or <https://arxiv.org/abs/quant-ph/0601075 its arXiv version>).
% 
% Each lattice site has spin-1/2 degree of freedoms, having two spin states, 
% $|{\uparrow_i}\rangle$ and $|{\downarrow_i}\rangle$. Each lattice site is decomposed 
% into four auxiliary sites, each of which corresponds to the bond direction: 
% left, up, down, and right. Each auxiliary site has the three-dimensional Hilbert 
% space: empty $|e\rangle$, spin-up $|{\Uparrow}\rangle$, and spin-down $|{\Downarrow}\rangle$. 
% The local physical states of the lattice site are given by the projection of 
% the product space of four auxiliary sites associated with the same lattice site.
% 
% The valence bond state is the entangled state of two auxiliary sites that 
% are associated with the bond connecting nearest-neighbor lattice sites. For 
% example, the valence bond state connecting a site $i$ and the other site $j$ 
% (that lies on the right of site $i$) is:
% 
% $$| E_{ir, jl} \rangle = | e_{i r}, e_{j l} \rangle + | {\Uparrow_{i r}}, 
% {\Downarrow_{j l}} \rangle - | {\Downarrow_{i r}}, {\Uparrow_{j l}} \rangle 
% ,$$
% 
% where $r$ and $l$ mean that the virtual sites for the right and left bonds, 
% respectively. The valence bond is represented by a rank-2 tensor:

clear
% left(in)-right(out) for horizontal valence bond,
% or up(in)-down(out) for vertical valence bond
VB = blkdiag(1,[0,1;-1,0]);
%% 
% Here we define the first, second, and third basis states as $|e\rangle$, $|{\Uparrow}\rangle$, 
% and $|{\Downarrow}\rangle$, respectively.
% 
% The projection onto local physical space is applied onto four auxiliary sites 
% at the same lattice site:
% 
% $$\hat{P}_i = | {\uparrow}_i \rangle \left( \langle {\Uparrow_{i l}}, e_{i 
% u}, e_{i d}, e_{i r} | + \langle e_{i l}, {\Uparrow_{i u}}, e_{i d}, e_{i r} 
% | + \langle e_{i l}, e_{i u}, {\Uparrow_{i d}}, e_{i r} | + \langle e_{i l}, 
% e_{i u}, e_{i d}, {\Uparrow_{i r}} | \right) \\ \qquad + | {\downarrow}_i \rangle 
% \left( \langle {\Downarrow_{i l}}, e_{i u}, e_{i d}, e_{i r} | + \langle e_{i 
% l}, {\Downarrow_{i u}}, e_{i d}, e_{i r} | + \langle e_{i l}, e_{i u}, {\Downarrow_{i 
% d}}, e_{i r} | + \langle e_{i l}, e_{i u}, e_{i d}, {\Downarrow_{i r}} | \right) 
% .$$
% 
% The numerical construction of $\hat{P}_i$ is done as:

P = zeros(3,3,2,3,3); % left(in)-up(in)-physical(in)-down(out)-right(out)

for it1 = (1:3) % left (in)
    for it2 = (1:3) % up (in)
        for it3 = (1:3) % down (out)
            for it4 = (1:3) % right (out)
                it_tot = [it1 it2 it3 it4];
                % spin-up in the physical space
                if (sum(it_tot == 1) == 3) && (sum(it_tot == 2) == 1)
                    P(it1,it2,1,it3,it4) = 1;
                end
                
                % spin-down in the physical space
                if (sum(it_tot == 1) == 3) && (sum(it_tot == 3) == 1)
                    P(it1,it2,2,it3,it4) = 1;
                end
            end
        end
    end
end
%% 
% Then the local tensor is obtained by contracting (i) the down leg of the projector 
% |P| with the up leg of vertical valence bond |VB| and (ii) the right leg of 
% the contraction result with the left leg of horizontal valence bond |VB|. (*Quick 
% exercise*: how about the left and up legs of |P|?)

A = contract(P,5,4,VB,2,1,[(1:3) 5 4]);
A = contract(A,5,5,VB,2,1);
%% Spin-spin correlation of the RVB state
% Let's compute the spin-spin correlation $\langle \hat{S}_{(i,j)}^z \hat{S}_{(i,j+1)}^z 
% \rangle$ for the nearest neighbors $(i, j)$ and $(i, j+1)$. We choose the row 
% and column indices $i$ and $j$ so that the sites lie at the center of the lattice. 
% With this choice, we can minimize the boundary effect.
% 
% Compute the spin-spin correlation for different system sizes.

[S,I] = getLocalSpace('Spin',1/2);
Sz = squeeze(S(:,3,:));

Nkeep = 30; % maximum bond dimension
Nrow = (6:2:10); % number of rows
Ncol = (11:20); % number of columns
Cvals = zeros(numel(Nrow),numel(Ncol)); % correlation

for itr = (1:numel(Nrow))
    for itc = (1:numel(Ncol))
        % location of spin operators
        rO = [round(Nrow(itr)/2),round((Ncol(itc)+1)/2)-1];
        rO = [rO;rO+[0 1]];
        
        val1 = normFinPEPS_Ex (A,Nrow(itr),Ncol(itc),Nkeep);
        val2 = normFinPEPS_Ex (A,Nrow(itr),Ncol(itc),Nkeep,Sz,rO);
        
        Cvals(itr,itc) = val2/val1;
    end
end
%%
figure;
hold on;
legs = cell(1,numel(Nrow));
for itr = (1:numel(Nrow))
    plot(Ncol,Cvals(itr,:),'LineWidth',1,'Marker','x');
    legs{itr} = ['# of rows = ',sprintf('%g',Nrow(itr))];
end
grid on;
set(gca,'LineWidth',1,'FontSize',13);
legend(legs(:),'Location','southeast');
xlabel('# of columns');
ylabel('Spin-spin correlation');
%% 
% The spin-spin correlation, as a function of the number of columns with fixed 
% number of rows, shows an oscillatory behaviour with period 4, and converges 
% slowly to a value between -0.095 and -0.1, in the limit of infinite columns. 
% A Monte Carlo calculation predicts the nearest-neighbour correlator $\langle 
% \hat{S}_{(i,j)}^z \hat{S}_{(i,j+1)}^z \rangle \simeq -0.0987$ in the thermodynamic 
% limit [<https://journals.aps.org/prb/abstract/10.1103/PhysRevB.82.180408 A. 
% F. Albuquerque and F. Alet, Phys. Rev. B *82*, 180408(R) (2010)> or <https://arxiv.org/abs/1009.1307v2 
% its arXiv version>].
%% Exercise (b): Different shape of square lattice
% In the above demonstration, we have considered the lattices whose rows are 
% longer than columns. How about the lattices whose columns are longer than rows? 
% Which result would be more accurate? Why?
%% Exercise (c): Toric code
% Compute the spin-spin correlator for the Kitaev's toric code.
% 
% (_Hint_: it should be always zero, since the action of two $\hat{S}^z$ operators 
% onto the ground state yields an excited energy eigenstate.) 
%% Exercise (d): Resonating AKLT loop (RAL) state
% Compute the spin-spin correlator for the RAL state. It is the generalization 
% of the RVB state: the RVB state is for the lattice of spin-1/2's, and the RAL 
% state is for the lattice of spin-1's, having three states $|+1_i\rangle$, $|0_i\rangle$, 
% and $|-1_i\rangle$. Similarly as in the PEPS construction of the RVB state, 
% each local tensor is given by the contraction of valence bond states and local 
% projectors. The valence bond state here is the same as one for the RVB state. 
% The local projector is different:
% 
% $$\hat{P}_i = | +1_i \rangle ( \langle e_{i l}, e_{i u}, {\Uparrow_{id}}, 
% {\Uparrow_{ir}} | + \langle e_{il} , {\Uparrow_{iu}} , e_{id} , {\Uparrow_{ir}} 
% | + \cdots ) \\ \quad\quad + | -1_i \rangle ( \langle e_{il} , e_{iu} , {\Downarrow_{id}} 
% , {\Downarrow_{ir}} | + \langle e_{il} , {\Downarrow_{iu}} , e_{id} , {\Downarrow_{ir}} 
% | + \cdots ) \\ \quad\quad + \frac{1}{\sqrt{2}} | 0_i \rangle ( \langle e_{il} 
% , e_{iu} , {\Uparrow_{id}} , {\Downarrow_{ir}} | +  \langle e_{il} , e_{iu} 
% , {\Downarrow_{id}} , {\Uparrow_{ir}} | + \langle e_{il} , {\Uparrow_{iu}} , 
% e_{id} , {\Downarrow_{ir}} | + \langle e_{il} , {\Downarrow_{iu}} , e_{id} , 
% {\Uparrow_{ir}} | + \cdots )$$
% 
% For details, refer to <https://journals.aps.org/prb/abstract/10.1103/PhysRevB.89.174411 
% Wei Li, Shuo Yang, Meng Cheng, Zheng-Xin Liu, and Hong-Hao Tu, Phys. Rev. B 
% *89*, 174411 (2014)>, <https://arxiv.org/abs/1312.3922 its arXiv version>, or 
% <https://www.theorie.physik.uni-muenchen.de/17ls_th_solidstate_en/publications/pdf/20140509_li.pdf 
% the local copy in our group website>.