function val = normFinPEPS_Ex (A,Nrow,Ncol,Nkeep,varargin)
% < Description >
%
% val = normFinPEPS_Ex (A,Nrow,Ncol,Nkeep [, O, rO]))
%
% Compute the norm of the finite PEPS which contists of uniform local
% tensors A on a Nrow * Ncol square lattice, with open boundary condition.
% We project the bond space of the open legs at the boundary (that are not
% connecting tensors) onto their first bases, to make the open legs be
% dummy legs of singleton dimensions.
% If the rank-2 local operator O acting on physical space and the location
% rO of acting O's are given, the local operators are inserted bewteen the
% ket and bra tensors.
% This functions contracts the tensor network, first by contracting along
% z-directions: Buliding transfer operators (also known as double tensors)
% that are the contraction of ket local tensor and bra local tensors. If
% local operator acts on some sites, then we considere different transfer
% operator, which is the contraction of ket, local operator, and bra
% tensors, for those sites. Then, we consider the first column of
% transfer operators as a matrix product state (MPS), and the other columns
% as matrix product operators (MPOs). We contract MPS and MPO, column by
% column. After contracting each column, we perform two sweeps (downward
% and upward, along column direction) to truncate bond spaces via the
% singular value decomposition.
%
% < Input >
% A : [rank-5 tensor] Local tensor in the bulk of the lattice. The leg
%       order is left-up-physical(local)-down-right. Organize the basis of
%       bond spaces of A so that the first bases are associated with the
%       vacuum space of dummy legs.
% Nrow, Ncol : [integer] Lattice size. Nrow means the number of rows, and
%       Ncol means the number of columns
% Nkeep : [integer] Maximum bond dimension. Set it as [] or Inf not to
%       truncate.
% O : [rank-2 tensor] (Optional) Local operator.
% rO : [n x 2 matrix] (Optional) Each row indicates the coordinate for a
%       site on which local operator O acts on. The first and second
%       columns of rO indicate row and column indices of the sites,
%       respectively.
% 
% < Output >
% val : [numeric] The norm of the tensor network.
%
% Written by S.Lee (Jul.06,2017)
% Updated by S.Lee (Jun.29,2019): Rewritten for SoSe 2019.


% % % for debugging
% try

tobj = tic2;

% parse input
if nargin > 4
    O = varargin{1};
    rO = varargin{2};
else
    O = [];
    rO = [];
end

% check the integrity of input and option
if ndims(A) > 5
    error('ERR: A should be rank-4.');
elseif size(A,1) ~= size(A,5)
    error('ERR: The left (1st) and right (5th) legs of A should have the same dimension.');
elseif size(A,2) ~= size(A,4)
    error('ERR: The up (2nd) and down (4th) legs of A should have the same dimension.');
end
if ~isempty(O)
    if ~ismatrix(O) || (size(O,1) ~= size(O,2))
        error('ERR: O should be square matrix.');
    elseif size(rO,2) ~= 2
        error('ERR: rO should be two-column matrix.');
    elseif size(O,1) ~= size(A,3)
        error('ERR: the local (3rd) leg of A should have the same dimension as the local operator O.');
    end
    
    for it1 = (1:size(rO,1))
        if (rO(it1,1) < 1) || (rO(it1,1) > Nrow) || (rO(it1,2) < 1) || (rO(it1,2) > Ncol)
            error('ERR: the location of operators ''rO'' should range from (1, 1) to (Nrow, Ncol).');
        end
    end
end

% display information
fprintf(['Finite PEPS contraction: Nrow = ',sprintf('%i',Nrow), ...
    ', Ncol = ',sprintf('%i',Ncol),', Nkeep = ',sprintf('%g',Nkeep),'\n']);
if ~isempty(O)
    fprintf('  Act local operator(s) onto (x, y) = ');
    for itr = (1:size(rO,1))
        fprintf('(%i, %i)',rO(itr,1),rO(itr,2));
        if itr < size(rO,1)
            fprintf(', ');
        else
            fprintf('\n');
        end
    end
end

% % transfer operator for the bulk, without local operator
T = contract(conj(A),5,3,A,5,3,reshape([(1:4);(5:8)],[1 8]));
% fuse the pair of legs in the same direction into a single fat leg
T = reshape(T,[size(T,1)*size(T,2),size(T,3)*size(T,4), ...
    size(T,5)*size(T,6),size(T,7)*size(T,8)]);
% NOTE: when we use QSpace, for example, then we need to use the isometries
% generated by getIdentity.

if ~isempty(O)
    % % % % TODO - Exercise (a)  (Start) % % % %
    % % transfer operator for the bulk, with local operator
    
    % fuse the pair of legs in the same direction into a single fat leg
    
    % % % % TODO - Exercise (a)  (End) % % % %
end

% total tensor network
Ttot = cell(Nrow,Ncol);
for itr = (1:Nrow) % itr: row index
    for itc = (1:Ncol) % itc: column index
        % leg order: left-up-down-right
        if ~isempty(O) && any(all(bsxfun(@eq, [itr itc], rO),2))
            Ttot{itr,itc} = TO;
        else
            Ttot{itr,itc} = T;
        end
        % project the open legs onto the first bases; the product of the
        % first basis of ket and the first basis of bra are also the first
        % basis in the product space
        if itr == 1
            % project up leg
            Ttot{itr,itc} = Ttot{itr,itc}(:,1,:,:);
        elseif itr == Nrow
            % project down leg
            Ttot{itr,itc} = Ttot{itr,itc}(:,:,1,:);
        end
        if itc == 1
            % project left leg
            Ttot{itr,itc} = Ttot{itr,itc}(1,:,:,:);
        elseif itc == Ncol
            % project right leg
            Ttot{itr,itc} = Ttot{itr,itc}(:,:,:,1);
        end 
    end
end

% The first column
M = Ttot(:,1);
for itr = (1:Nrow)
    % each tensor is effectively rank-3; order the legs to be up-down-right
    % (-left). The left legs are dummy legs, so will be neglected.
    M{itr} = permute(M{itr},[2 3 4 1]);
end

% overall prefactor of norm; cumulatively multiply the norm of the rank-4
% tensors (that are to be SVD-ed)
Nfac = 1;

for itc = (2:Ncol)
    % % % % TODO - Exercise (a)  (Start) % % % %
    % contract column by column: contract "MPS tensors" with the tensors in
    % the next column
    
    % top-to-bottom sweep of truncating bonds
    

    % bottom-to-top sweep of truncating bonds
    

    % % % % TODO - Exercise (a)  (End) % % % %
end

val = M{1}; % up-down-right (here up and right leg are dummy, so M{1} is row vector)
for itr = (2:Nrow)
    val = val*M{itr};
end
val = val*Nfac;

toc2(tobj,'-v');

% % % for debugging
% catch e
%     disp(getReport(e))
%     disp('Let''s Debug!');
%     keyboard;
% end

end